@@ -1,4 +1,4 @@
-package Pcore::PgSQL v0.15.1;
+package Pcore::PgSQL v0.15.2;
 
 use Pcore -dist, -class;
 
@@ -16,14 +16,21 @@ sub run ( $self, $cb ) {
     if ( $self->is_empty ) {
         my $pwfile = "$self->{data_dir}/pgsql-password.txt";
 
-        my $superuser_password = defined $ENV{PGSQL_POSTGRES_PASSWORD} ? $ENV{PGSQL_POSTGRES_PASSWORD} : P->random->bytes_hex(32);
+        my $superuser_password;
+
+        if ( defined $ENV{PGSQL_POSTGRES_PASSWORD} ) {
+            $superuser_password = $ENV{PGSQL_POSTGRES_PASSWORD};
+        }
+        else {
+            $superuser_password = P->random->bytes_hex(32);
+
+            say "GENERATED POSTGRES PASSWORD: $superuser_password\n";
+        }
 
         P->file->write_bin( $pwfile, $superuser_password );
 
         chown $uid, $uid, $pwfile or die;
 
-        say "GENERATED POSTGRES PASSWORD: $superuser_password\n";
-
         my $res = P->pm->run_proc( [ 'su', 'postgres', '-c', "$ENV{POSTGRES_HOME}/bin/initdb --encoding UTF8 --no-locale -U postgres --pwfile $pwfile -D $db_dir" ] );
 
         unlink $pwfile or 1;
